# -*- coding: utf-8 -*-
import cmath
from typing import Dict, List, Tuple

import numpy as np
import scipy.optimize as so
import scipy.special as sp
from riip import Material

import pymwm
from pymwm.cutoff import Cutoff
from pymwm.waveguide import Sampling


class Samples(Sampling):
    """A class defining samples of phase constants of coaxial waveguide
    modes.

    Attributes:
        ri: A float indicating inner radius [um].
        r: A float indicating outer radius [um].
    """

    def __init__(
        self, size: float, fill: Material, clad: Material, params: Dict, size2: float
    ):
        """Init Samples class.

        Args:
            size: A float indicating the outer radius [um]
            fill: An instance of Material class for the core
            clad: An instance of Material class for the clad
            params: A dict whose keys and values are as follows:
                'wl_max': A float indicating the maximum wavelength [um]
                    (default: 5.0)
                'wl_min': A float indicating the minimum wavelength [um]
                    (default: 0.4)
                'wl_imag': A float indicating the minimum value of
                    abs(c / f_imag) [um] where f_imag is the imaginary part of
                    the frequency. (default: 5.0)
                'dw': A float indicating frequency interval
                    [rad c / 1um]=[2.99792458e14 rad / s] (default: 1 / 64).
                'num_n': An integer indicating the number of orders of modes.
                'num_m': An integer indicating the number of modes in each
                    order and polarization.
            size2: A float indicating the inner radius [um]
        """
        self.r = size
        self.ri = size2
        super().__init__(size, fill, clad, params, size2)
        self.r_ratio = self.ri / self.r
        num_n = self.params["num_n"]
        num_m = self.params["num_m"]
        co = Cutoff(num_n, num_m)
        self.co_list = []
        for n in range(num_n):
            co_per_n = []
            for pol in ["M", "E"]:
                for m in range(1, num_m + 1):
                    alpha = (pol, n, m)
                    co_per_n.append(co(alpha, self.r_ratio))
            self.co_list.append(np.array(co_per_n))

    @property
    def shape(self):
        return "coax"

    @property
    def num_all(self):
        return self.params["num_n"] * (2 * self.params["num_m"] + 1)

    def beta2_pec(self, w, n):
        """Return squares of phase constants for a PEC waveguide

        Args:
            w: A complex indicating the angular frequency
            n: A integer indicating the order of the modes.
        Returns:
            h2s: A 1D array indicating squares of phase constants, whose first
                num_m+1 elements are for TM-like modes and the rest are for
                TE-like modes.
        """
        return self.fill(w) * w ** 2 - self.co_list[n] ** 2 / self.r ** 2

    def u(self, h2: complex, w: complex, e2: complex) -> complex:
        # return cmath.sqrt(e2 * w ** 2 - h2) * self.r
        return (1 - 1j) * cmath.sqrt(0.5j * (h2 - e2 * w ** 2)) * self.ri

    def v(self, h2: complex, w: complex, e1: complex) -> complex:
        return (1 + 1j) * cmath.sqrt(-0.5j * (e1 * w ** 2 - h2)) * self.ri

    def x(self, h2: complex, w: complex, e1: complex) -> complex:
        return (1 + 1j) * cmath.sqrt(-0.5j * (e1 * w ** 2 - h2)) * self.r

    def y(self, h2: complex, w: complex, e2: complex) -> complex:
        return (1 - 1j) * cmath.sqrt(0.5j * (h2 - e2 * w ** 2)) * self.r

    def eig_eq(
        self, h2: complex, w: complex, pol: str, n: int, e1: complex, e2: complex
    ):
        """Return the value of the characteristic equation

        Args:
            h2: The square of the phase constant.
            w: The angular frequency
            pol: The polarization
            n: The order of the modes
            e1: The permittivity of the core
            e2: The permittivity of the clad.
        Returns:
            val: A complex indicating the left-hand value of the characteristic
                equation.
        """
        h2c = h2.real + 1j * h2.imag
        w2 = w ** 2
        u = self.u(h2c, w, e2)
        v = self.v(h2c, w, e1)
        x = self.x(h2c, w, e1)
        y = self.y(h2c, w, e2)

        def F(f_name, z):
            sign = 1 if f_name == "iv" else -1
            func = eval("sp." + f_name)
            f = func(n, z)
            fp = sign * func(n + 1, z) + n / z * f
            return f, fp / f

        jv, Fjv = F("jv", v)
        jx, Fjx = F("jv", x)
        yv, Fyv = F("yv", v)
        yx, Fyx = F("yv", x)
        iu, Fiu = F("iv", u)
        ky, Fky = F("kv", y)

        nuv = n * (u / v + v / u)
        nxy = n * (x / y + y / x)

        a11 = (v * Fiu + u * Fjv) / yv
        a12 = v * Fiu + u * Fyv
        a13 = nuv / yv
        a14 = nuv

        a21 = h2c * nuv / e2 / yv
        a22 = h2c * nuv / e2
        a23 = w2 * (v * Fiu + e1 / e2 * u * Fjv) / yv
        a24 = w2 * (v * Fiu + e1 / e2 * u * Fyv)

        a31 = (x * Fky + y * Fjx) / yx
        a32 = (x * Fky + y * Fyx) * jv / jx
        a33 = nxy / yx
        a34 = nxy * jv / jx

        a41 = h2c * nxy / e2 / yx
        a42 = h2c * nxy / e2 * jv / jx
        a43 = w ** 2 * (x * Fky + e1 / e2 * y * Fjx) / yx
        a44 = w ** 2 * (x * Fky + e1 / e2 * y * Fyx) * jv / jx

        return np.linalg.det(
            np.array(
                [
                    [a11, a12, a13, a14],
                    [a21, a22, a23, a24],
                    [a31, a32, a33, a34],
                    [a41, a42, a43, a44],
                ]
            )
        )

    def beta2(self, w, n, e1, e2, xis):
        """Return roots and convergences of the characteristic equation

        Args:
            w: A complex indicating the angular frequency.
            n: A integer indicating the order of the mode
            e1: A complex indicating the permittiity of tha core.
            e2: A complex indicating the permittivity of tha clad.
            xis: A complex indicating the initial approximations for the roots
                whose number of elements is 2*num_m+1.
        Returns:
            xs: A 1D array indicating the roots, whose length is 2*num_m+1.
            success: A 1D array indicating the convergence information for xs.
        """
        if self.clad.label == "PEC":
            xs = self.beta2_pec(w, n)
            success = np.ones_like(xs, dtype=bool)
            return xs, success
        from scipy.optimize import root

        num_m = self.params["num_m"]
        roots = []
        vals = []
        success = []

        def func(h2vec, *pars):
            h2 = h2vec[0] + h2vec[1] * 1j
            f = self.eig_eq(h2, *pars)
            prod_denom = 1.0
            for h2_0 in roots:
                denom = h2 - h2_0
                while abs(denom) < 1e-14:
                    denom += 1.0e-14
                prod_denom *= 1.0 / denom
            f *= prod_denom
            f_array = np.array([f.real, f.imag])
            return f_array

        for i, xi in enumerate(xis):
            if i < num_m:
                args = (w, "M", n, e1, e2)
            else:
                args = (w, "E", n, e1, e2)
            # result = root(func, (xi.real, xi.imag), args=args, jac=True,
            #               method='hybr', options={'xtol': 1.0e-9})
            result = root(
                func,
                np.array([xi.real, xi.imag]),
                args=args,
                jac=False,
                method="hybr",
                options={"xtol": 1.0e-9},
            )
            x = result.x[0] + result.x[1] * 1j
            v = self.v(x, w, e2)
            if result.success:
                roots.append(x)
            # if v.real > 0.0 and v.real > abs(v.imag):
            if v.real > 0.0:
                success.append(result.success)
            else:
                success.append(False)
            vals.append(x)
        return np.array(vals), success

    def beta2_w_min(self, n):
        """Return roots and convergences of the characteristic equation at
            the lowest angular frequency, ws[0].

        Args:
            n: A integer indicating the order of the mode
        Returns:
            xs: A 1D array indicating the roots, whose length is 2*num_m+1.
            success: A 1D array indicating the convergence information for xs.
        """
        w_0 = 0.1
        if self.clad.label == "PEC":
            xs = self.beta2_pec(self.ws[0], n)
            success = np.ones_like(xs, dtype=bool)
            return xs, success
        e1 = self.fill(w_0)
        e2_0 = -5.0e5 + self.clad(w_0).imag * 1j
        de2 = (self.clad(w_0) - e2_0) / 1000
        xis = xs = self.beta2_pec(w_0, n)
        success = np.ones_like(xs, dtype=bool)
        for i in range(1001):
            e2 = e2_0 + de2 * i
            xs, success = self.beta2(w_0, n, e1, e2, xis)
            for _, ok in enumerate(success):
                if not ok:
                    xs[_] = xis[_]
            xis = xs
        dw = (self.ws[0] - w_0) / 1000
        for i in range(1001):
            w = w_0 + dw * i
            e1 = self.fill(w)
            e2 = self.clad(w)
            xs, success = self.beta2(w, n, e1, e2, xis)
            for _, ok in enumerate(success):
                if not ok:
                    xs[_] = xis[_]
            xis = xs
        return xs, success

    @staticmethod
    def beta_from_beta2(x):
        return (1 + 1j) * np.sqrt(-0.5j * x)
        # val = np.sqrt(x)
        # if ((abs(val.real) > abs(val.imag) and val.real < 0) or
        #    (abs(val.real) < abs(val.imag) and val.imag < 0)):
        #     val *= -1
        # return val

    def __call__(self, n: int):
        """Return a dict of the roots of the characteristic equation

        Args:
            n: A integer indicating the order of the mode
        Returns:
            betas: A dict containing arrays of roots, whose key is as follows:
                (pol, n, m):
                    pol: 'E' or 'M' indicating the polarization.
                    n: A integer indicating the order of the mode.
                    m: A integer indicating the ordinal of the mode in the same
                        order.
            convs: A dict containing the convergence information for betas,
                whose key is the same as above.
        """
        num_m = self.params["num_m"]
        xs_array = np.zeros((len(self.ws), len(self.wis), 2 * num_m + 1), dtype=complex)
        success_array = np.zeros(
            (len(self.ws), len(self.wis), 2 * num_m + 1), dtype=bool
        )
        iwr = iwi = 0
        wi = self.wis[iwi]
        xis, success = self.beta2_w_min(n)
        xs_array[iwr, iwi] = xis
        success_array[iwr, iwi] = success
        for iwr in range(1, len(self.ws)):
            wr = self.ws[iwr]
            w = wr + 1j * wi
            e1 = self.fill(w)
            e2 = self.clad(w)
            xs, success = self.beta2(w, n, e1, e2, xis)
            for i, ok in enumerate(success):
                # if ok:
                #     if abs(xs[i] - xis[i]) > max(0.1 * abs(xis[i]), 5.0):
                #         success[i] = False
                #         xs[i] = xis[i]
                # else:
                if not ok:
                    xs[i] = xis[i]
            xs_array[iwr, iwi] = xs
            success_array[iwr, iwi] = success
            xis = xs
            # print(iwr, iwi, success)
        for iwi in range(1, len(self.wis)):
            wi = self.wis[iwi]
            for iwr in range(len(self.ws)):
                wr = self.ws[iwr]
                w = wr + 1j * wi
                e1 = self.fill(w)
                e2 = self.clad(w)
                if iwr == 0:
                    xis = xs_array[iwr, iwi - 1]
                else:
                    xis = (
                        xs_array[iwr, iwi - 1]
                        + xs_array[iwr - 1, iwi]
                        - xs_array[iwr - 1, iwi - 1]
                    )
                xs, success = self.beta2(w, n, e1, e2, xis)
                for i, ok in enumerate(success):
                    # if ok:
                    #     if abs(xs[i] - xis[i]) > max(
                    #             0.1 * abs(xis[i]), 5.0):
                    #         success[i] = False
                    #         xs[i] = xis[i]
                    # else:
                    if not ok:
                        xs[i] = xis[i]
                xs_array[iwr, iwi] = xs
                success_array[iwr, iwi] = success
                # print(iwr, iwi, success)
        return xs_array, success_array

    def betas_convs(self, xs_success_list):
        num_n = self.params["num_n"]
        num_m = self.params["num_m"]
        betas = {}
        convs = {}
        for n in range(num_n):
            xs_array, success_array = xs_success_list[n]
            for m in range(1, num_m + 2):
                betas[("M", n, m)] = np.zeros(
                    (len(self.ws), len(self.wis)), dtype=complex
                )
                convs[("M", n, m)] = np.zeros((len(self.ws), len(self.wis)), dtype=bool)
            for m in range(1, num_m + 1):
                betas[("E", n, m)] = np.zeros(
                    (len(self.ws), len(self.wis)), dtype=complex
                )
                convs[("E", n, m)] = np.zeros((len(self.ws), len(self.wis)), dtype=bool)
            for iwi in range(len(self.wis)):
                for iwr in range(len(self.ws)):
                    w = self.ws[iwr] + 1j * self.wis[iwi]
                    e2 = self.clad(w)
                    for i in range(num_m + 1):
                        x = xs_array[iwr, iwi][i]
                        v = self.v(x, w, e2)
                        betas[("M", n, i + 1)][iwr, iwi] = self.beta_from_beta2(x)
                        convs[("M", n, i + 1)][iwr, iwi] = (
                            success_array[iwr, iwi][i]
                            if v.real > abs(v.imag)
                            else False
                        )
                    for i in range(num_m):
                        x = xs_array[iwr, iwi][i + num_m + 1]
                        v = self.v(x, w, e2)
                        betas[("E", n, i + 1)][iwr, iwi] = self.beta_from_beta2(x)
                        convs[("E", n, i + 1)][iwr, iwi] = (
                            success_array[iwr, iwi][i + num_m + 1]
                            if v.real > abs(v.imag)
                            else False
                        )
        return betas, convs


class SamplesLowLoss(Samples):
    """A class defining samples of phase constants of cylindrical waveguide
    modes in a virtual low-loss clad waveguide by subclassing the Samples
    class.

    Attributes:
        r: A float indicating the outer radius [um]
        ri: A float indicating the inner radius [um]
        fill: An instance of Material class for the core
        clad: An instance of Material class for the clad
        params: A dict whose keys and values are as follows:
            'wl_max': A float indicating the maximum wavelength [um]
            'wl_min': A float indicating the minimum wavelength [um]
            'wl_imag': A float indicating the minimum value of
                abs(c / f_imag) [um] where f_imag is the imaginary part of
                the frequency.
            'dw': A float indicating frequency interval
                [rad * c / 1um]=[2.99792458e14 rad / s].
            'num_n': An integer indicating the number of orders of modes.
    """

    def __init__(self, size, fill, clad, params, size2):
        """Init Samples class.

        Args:
            size: A float indicating the outer radius [um]
            r: A float indicating the radius of the circular cross section [um]
            fill: An instance of Material class for the core
            clad: An instance of Material class for the clad
            params: A dict whose keys and values are as follows:
                'wl_max': A float indicating the maximum wavelength [um]
                    (default: 5.0)
                'wl_min': A float indicating the minimum wavelength [um]
                    (default: 0.4)
                'wl_imag': A float indicating the minimum value of
                    abs(c / f_imag) [um] where f_imag is the imaginary part of
                    the frequency. (default: 5.0)
                'dw': A float indicating frequency interval
                    [rad c / 1um]=[2.99792458e14 rad / s] (default: 1 / 64).
                'num_n': An integer indicating the number of orders of modes.
                'num_m': An integer indicating the number of modes in each
                    order and polarization.
            size2: A float indicating the inner radius [um]
        """
        super(SamplesLowLoss, self).__init__(size, fill, clad, params, size2)

    def __call__(self, arg: Tuple[int, int, List[np.ndarray]]):
        """Return a dict of the roots of the characteristic equation

        Args:
            arg: (iwr, iwi, xis_list)
                iwr: The ordinal of the Re(w).
                iwi: The ordinal of the Im(w).
                xis_list: The initial guess of roots whose length is 2*num_m+1
        Returns:
            xs_list: A list of num_n 1D arrays indicating the roots, whose
                length is 2*num_m+1
            success_list: A list of num_n 1D arrays indicating the convergence
                information for xs, whose length is 2*num_m+1
        """
        iwr, iwi, xis_list = arg
        im_factor = self.clad.im_factor
        self.clad.im_factor = 1.0
        wr = self.ws[iwr]
        wi = self.wis[iwi]
        w = wr + 1j * wi
        e1 = self.fill(w)
        xs_list = []
        success_list = []
        for n, x0s in enumerate(xis_list):
            xis = xs = x0s
            success = np.ones_like(xs, dtype=bool)
            for i in range(1, 16):
                self.clad.im_factor = 0.7 ** i
                if i == 15 or self.clad.im_factor < im_factor:
                    self.clad.im_factor = im_factor
                e2 = self.clad(w)
                xs, success = self.beta2(w, n, e1, e2, xis)
                for _, ok in enumerate(success):
                    if not ok:
                        xs[_] = xis[_]
                xis = xs
            xs_list.append(xs)
            success_list.append(success)
        return xs_list, success_list

    def betas_convs(self, xs_success_list):
        num_iwr = len(self.ws)
        num_iwi = len(self.wis)
        num_n = self.params["num_n"]
        num_m = self.params["num_m"]
        betas = {}
        convs = {}
        for n in range(num_n):
            for m in range(1, num_m + 1):
                betas[("M", n, m)] = np.zeros(
                    (len(self.ws), len(self.wis)), dtype=complex
                )
                convs[("M", n, m)] = np.zeros((len(self.ws), len(self.wis)), dtype=bool)
            for m in range(1, num_m + 1):
                betas[("E", n, m)] = np.zeros(
                    (len(self.ws), len(self.wis)), dtype=complex
                )
                convs[("E", n, m)] = np.zeros((len(self.ws), len(self.wis)), dtype=bool)
        for iwr in range(num_iwr):
            for iwi in range(num_iwi):
                j = iwr * num_iwi + iwi
                w = self.ws[iwr] + 1j * self.wis[iwi]
                e2 = self.clad(w)
                for n in range(num_n):
                    for i in range(num_m):
                        x = xs_success_list[j][0][n][i]
                        v = self.v(x, w, e2)
                        betas[("M", n, i + 1)][iwr, iwi] = self.beta_from_beta2(x)
                        convs[("M", n, i + 1)][iwr, iwi] = (
                            xs_success_list[j][1][n][i]
                            if v.real > abs(v.imag)
                            else False
                        )
                    for i in range(num_m):
                        x = xs_success_list[j][0][n][i + num_m + 1]
                        v = self.v(x, w, e2)
                        betas[("E", n, i + 1)][iwr, iwi] = self.beta_from_beta2(x)
                        convs[("E", n, i + 1)][iwr, iwi] = (
                            xs_success_list[j][1][n][i + num_m + 1]
                            if v.real > abs(v.imag)
                            else False
                        )
        return betas, convs



# -*- coding: utf-8 -*-
# cython: profile=False
from . cimport cabs, ccos, cdouble, creal, csin, csqrt, ctan, sqrt


cdef coefs_pec_C(long *s_all, long *n_all, int num_n_all,
              double r, cdouble *As, cdouble *Bs)

cdef cdouble csinc(cdouble x) nogil

cdef void coefs_C(
    cdouble *hs, cdouble w, long *s_all, long *n_all,
    int num_n_all, double r, cdouble e1, cdouble e2,
    cdouble *As, cdouble *Bs) nogil



import os

import numpy as np
from Cython.Distutils import build_ext
from setuptools import Extension, find_packages, setup

ext_modules = []
for name in ["cylinder", "slit", "bessel", "cutoff"]:
    basename = os.path.join("src", "pymwm", "utils", f"{name}_utils")
    e = Extension(
        f"pymwm.utils.{name}_utils",
        sources=[basename + ".pyx"],
        depends=[basename + ".pxd"],
        include_dirs=[np.get_include(), "."],
        language="c++",
    )
    e.cython_directives = {"language_level": "3"}
    ext_modules.append(e)

setup(
    name="pymwm",
    version="0.2.9",
    url="https://github.com/mnishida/PyMWM",
    license="MIT",
    author="Munehiro Nishida",
    author_email="mnishida@hiroshima-u.ac.jp",
    description="A metallic waveguide mode solver",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    zip_safe=False,
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    setup_requires=["Cython", "numpy", "scipy"],
    install_requires=[line.strip() for line in open("requirements.txt").readlines()],
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
    ],
    keywords="metallic waveguide mode, electromagnetism",
    ext_modules=ext_modules,
    cmdclass={"build_ext": build_ext},
    entry_points={
        "console_scripts": ["update_pymwm_h5 = pymwm.bin.update_pymwm_h5:main"]
    },
)



# -*- coding: utf-8 -*-
# cython: profile=False
import numpy as np

cimport cython
cimport numpy as np


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void jk_to_coefs(int n, int pol, cdouble h,
                      cdouble u, cdouble jnu, cdouble jnpu,
                      cdouble v, cdouble knv, cdouble knpv,
                      cdouble w, double r,
                      cdouble e1, cdouble e2,
                      cdouble ab[2]) nogil:
    cdef:
        double norm
        cdouble c
    c = - n * (u * u + v * v) * jnu * knv / (u * v)
    if pol == 0:
        c *= h * h / (w * w)
        c /= (e1 * jnpu * v * knv + e2 * knpv * u * jnu)
        # norm = sqrt(1.0 + cabs(c) * cabs(c))
        # ab[0] = 1.0 / norm
        # ab[1] = c / norm
        ab[0] = 1.0
        ab[1] = c
    else:
        c /= (jnpu * v * knv + knpv * u * jnu)
        # norm = sqrt(1.0 + cabs(c) * cabs(c))
        # ab[1] = 1.0 / norm
        # ab[0] = c / norm
        ab[1] = 1.0
        ab[0] = c


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef cdouble upart_diag(int n, cdouble uc, cdouble jnuc, cdouble jnpuc,
                        cdouble u, cdouble jnu, cdouble jnpu) nogil:
    cdef:
        int n2 = n * n
        int sign
        cdouble u2, jnu2, jnpu2, u0, jnu0, jnpu0
        cdouble vals[2]
    if cabs(uc - u) < 1e-10:
        u0 = (u + uc) / 2
        sign = 1
        jv_jvp(n, u0, vals)
        jnu0 = vals[0]
        jnpu0 = vals[1]
        u2 = u0 * u0
        jnu2 = jnu0 * jnu0
        jnpu2 = jnpu0 * jnpu0
        return (sign *
                (jnu0 * jnpu0 / u0 + (jnpu2 + (1.0 - n2 / u2) * jnu2) / 2.0))
    elif cabs(uc + u) < 1e-10:
        u0 = (u - uc) / 2
        sign = 1 - ((n - 1) % 2) * 2
        jv_jvp(n, u0, vals)
        jnu0 = vals[0]
        jnpu0 = vals[1]
        u2 = u0 * u0
        jnu2 = jnu0 * jnu0
        jnpu2 = jnpu0 * jnpu0
        return (sign *
                (jnu0 * jnpu0 / u0 + (jnpu2 + (1.0 - n2 / u2) * jnu2) / 2.0))
    else:
        u2 = u * u
        jnu2 = jnu * jnu
        jnpu2 = jnpu * jnpu
        return (uc * jnuc * jnpu - u * jnu * jnpuc) / (uc * uc - u2)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline cdouble upart_off(int n, cdouble uc, cdouble jnuc,
                              cdouble u, cdouble jnu) nogil:
    return n * (jnuc * jnu) / (uc * u)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef cdouble vpart_diag(int n, cdouble vc, cdouble knvc, cdouble knpvc,
                        cdouble v, cdouble knv, cdouble knpv) nogil:
    cdef:
        int n2 = n * n
        int sign
        cdouble v2, knv2, knpv2, v0, knv0, knpv0
        cdouble vals[2]
    if cabs(vc - v) < 1e-10:
        v0 = (v + vc) / 2
        sign = 1
        kv_kvp(n, v0, vals)
        knv0 = vals[0]
        knpv0 = vals[1]
        v2 = v0 * v0
        knv2 = knv0 * knv0
        knpv2 = knpv0 * knpv0
        return sign * (
            knv0 * knpv0 / v0 +
            (knpv2 - (1.0 + n2 / v2) * knv2) / 2.0)
    elif cabs(vc + v) < 1e-10:
        v0 = (v - vc) / 2
        sign = 1 - ((n - 1) % 2) * 2
        kv_kvp(n, v0, vals)
        knv0 = vals[0]
        knpv0 = vals[1]
        v2 = v0 * v0
        knv2 = knv0 * knv0
        knpv2 = knpv0 * knpv0
        return sign * (
            knv0 * knpv0 / v0 +
            (knpv2 - (1.0 + n2 / v2) * knv2) / 2.0)
    else:
        v2 = v * v
        knv2 = knv * knv
        knpv2 = knpv * knpv
        return (vc * knvc * knpv -
                v * knv * knpvc) / (vc * vc - v2)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline cdouble vpart_off(int n, cdouble vc, cdouble knvc,
                              cdouble v, cdouble knv) nogil:
    return n * (knvc * knv) / (vc * v)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def coefs_cython(object hole, cdouble[::1] hs, cdouble w):
    cdef:
        long[::1] s_all = hole.s_all
        long[::1] n_all = hole.n_all
        long[::1] m_all = hole.m_all
    As_array = np.empty(hole.num_n_all, dtype=complex)
    Bs_array = np.empty(hole.num_n_all, dtype=complex)
    cdef:
        cdouble[::1] As = As_array
        cdouble[::1] Bs = Bs_array
        cdouble e1 = hole.fill(w)
        cdouble e2 = hole.clad(w)
    if creal(e2) < -1e6:
        coefs_pec_C(
            &s_all[0], &n_all[0], &m_all[0], hole.num_n_all,
            hole.r, &As[0], &Bs[0])
    else:
        coefs_C(
            &hs[0], w, &s_all[0], &n_all[0], &m_all[0], hole.num_n_all,
            hole.r, e1, e2, &As[0], &Bs[0])
    return As_array, Bs_array


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void coefs_pec_C(long *s_all, long *n_all, long  *m_all, int num_n_all,
              double r, cdouble *As, cdouble *Bs):
    cdef:
        int i, n, m, en
        double u, norm, jnu, jnpu
    from scipy.special import jn_zeros, jnp_zeros, jv, jvp
    for i in range(num_n_all):
        n = n_all[i]
        m = m_all[i]
        en = 1 if n == 0 else 2
        if s_all[i] == 0:
            u = jnp_zeros(n, m)[m - 1]
            jnu = jv(n, u)
            norm = sqrt(M_PI * r * r / en  *
                         (1 - n * n / (u * u)) * jnu * jnu)
            As[i] = 1.0 / norm
            Bs[i] = 0.0
        else:
            u = jn_zeros(n, m)[m - 1]
            jnpu = jvp(n, u)
            norm = sqrt(M_PI * r * r / en * jnpu * jnpu)
            As[i] = 0.0
            Bs[i] = 1.0 / norm


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void coefs_C(
    cdouble *hs, cdouble w, long *s_all, long *n_all, long  *m_all,
    int num_n_all, double r, cdouble e1, cdouble e2,
    cdouble *As, cdouble *Bs) nogil:
    cdef:
        int i, s, n, m, en
        cdouble norm
        cdouble h, u, v, jnu, jnpu, knv, knpv, a, b
        cdouble ab[2]
        cdouble vals[2]
        cdouble val_u, val_v, ud, uod, vd, vod
    for i in range(num_n_all):
        n = n_all[i]
        en = 1 if n == 0 else 2
        h = hs[i]
        s = s_all[i]
        u = (1 + 1j) * csqrt(-0.5j * (e1 * w * w - h * h)) * r
        jv_jvp(n, u, vals)
        jnu = vals[0]
        jnpu = vals[1]
        v = (1 - 1j) * csqrt(0.5j * (- e2 * w * w + h * h)) * r
        kv_kvp(n, v, vals)
        knv = vals[0]
        knpv = vals[1]
        jk_to_coefs(n, s, h, u, jnu, jnpu, v, knv, knpv, w, r, e1, e2, ab)
        a = ab[0]
        b = ab[1]
        val_u = 2 * M_PI * r * r / en
        val_v = (u * jnu) / (v * knv)
        val_v *= val_u * val_v
        ud = upart_diag(n, u, jnu, jnpu, u, jnu, jnpu)
        vd = vpart_diag(n, v, knv, knpv, v, knv, knpv)
        uod = upart_off(n, u, jnu, u, jnu)
        vod = vpart_off(n, v, knv, v, knv)
        norm = csqrt(val_u * (
            a * (a * ud + b * uod) + b * (b * ud + a * uod)) - val_v * (
            a * (a * vd + b * vod) + b * (b * vd + a * vod)))
        As[i] = a / norm
        Bs[i] = b / norm


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def ABY_cython(cdouble w, double r, long[::1] s_all, long[::1] n_all,
               long[::1] m_all, cdouble[::1] hs, cdouble e1, cdouble e2,
               double[:, :, ::1] u_pec, double[:, :, ::1] jnu_pec,
               double[:, :, ::1] jnpu_pec):
    cdef:
        int i, s, n, m, en
        int num_n_all = n_all.shape[0]
        double up, norm, jnup, jnpup
        cdouble h, u, v, jnu, jnpu, knv, knpv, a, b
        cdouble uc, vc, jnuc, jnpuc, knvc, knpvc, ac, bc
        cdouble ab[2]
        cdouble vals[2]
        cdouble val_u, val_v, ud, uod, vd, vod
    As_array = np.empty(num_n_all, dtype=complex)
    Bs_array = np.empty(num_n_all, dtype=complex)
    Ys_array = np.empty(num_n_all, dtype=complex)
    cdef:
        cdouble[:] As = As_array
        cdouble[:] Bs = Bs_array
        cdouble[:] Ys = Ys_array
    if creal(e2) < -1e6:
        for i in range(num_n_all):
            s = s_all[i]
            n = n_all[i]
            m = m_all[i]
            en = 1 if n == 0 else 2
            up = u_pec[s, n, m - 1]
            jnup = jnu_pec[s, n, m - 1]
            jnpup = jnpu_pec[s, n, m - 1]
            if s_all[i] == 0:
                norm = sqrt(M_PI * r * r / en  *
                             (1 - n * n / (up * up)) * jnup * jnup)
                As[i] = 1.0 / norm
                Bs[i] = 0.0
                Ys[i] = hs[i] / w
            else:
                norm = sqrt(M_PI * r * r / en * jnpup * jnpup)
                As[i] = 0.0
                Bs[i] = 1.0 / norm
                Ys[i] = e1 * w / hs[i]
        return As_array, Bs_array, Ys_array
    coefs_C(
        &hs[0], w, &s_all[0], &n_all[0], &m_all[0], num_n_all,
        r, e1, e2, &As[0], &Bs[0])
    for i in range(num_n_all):
        s = s_all[i]
        n = n_all[i]
        en = 1 if n == 0 else 2
        h = hs[i]
        u = (1 + 1j) * csqrt(-0.5j * (e1 * w * w - h * h)) * r
        jv_jvp(n, u, vals)
        jnu = vals[0]
        jnpu = vals[1]
        v = (1 - 1j) * csqrt(0.5j * (- e2 * w * w + h * h)) * r
        kv_kvp(n, v, vals)
        knv = vals[0]
        knpv = vals[1]
        val_u = 2 * M_PI * r * r / en
        val_v = val_u * ((u * jnu) / (v * knv)) ** 2
        ud = upart_diag(n, u, jnu, jnpu, u, jnu, jnpu)
        vd = vpart_diag(n, v, knv, knpv, v, knv, knpv)
        uod = upart_off(n, u, jnu, u, jnu)
        vod = vpart_off(n, v, knv, v, knv)
        a = As[i]
        b = Bs[i]
        Ys[i] = (val_u * (h / w * a * (a * ud + b * uod) +
                          e1 * w / h * b * (b * ud + a * uod)) -
                 val_v * (h / w * a * (a * vd + b * vod) +
                          e2 * w / h * b * (b * vd + a * vod)))
    return As_array, Bs_array, Ys_array


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def uvABY_cython(cdouble w, double r, long[::1] s_all, long[::1] n_all,
                 long[::1] m_all, cdouble[::1] hs, cdouble e1, cdouble e2,
                 double[:, :, ::1] u_pec, double[:, :, ::1] jnu_pec,
                 double[:, :, ::1] jnpu_pec):
    cdef:
        int i, s, n, m, en
        int num_n_all = n_all.shape[0]
        double up, norm, jnup, jnpup
        cdouble h, u, v, jnu, jnpu, knv, knpv, a, b
        cdouble uc, vc, jnuc, jnpuc, knvc, knpvc, ac, bc
        cdouble ab[2]
        cdouble vals[2]
        cdouble val_u, val_v, ud, uod, vd, vod
    us_array = np.empty(num_n_all, dtype=complex)
    vs_array = np.empty(num_n_all, dtype=complex)
    As_array = np.empty(num_n_all, dtype=complex)
    Bs_array = np.empty(num_n_all, dtype=complex)
    Ys_array = np.empty(num_n_all, dtype=complex)
    cdef:
        cdouble[:] us = us_array
        cdouble[:] vs = vs_array
        cdouble[:] As = As_array
        cdouble[:] Bs = Bs_array
        cdouble[:] Ys = Ys_array
    if creal(e2) < -1e6:
        for i in range(num_n_all):
            s = s_all[i]
            n = n_all[i]
            m = m_all[i]
            en = 1 if n == 0 else 2
            up = u_pec[s, n, m - 1]
            jnup = jnu_pec[s, n, m - 1]
            jnpup = jnpu_pec[s, n, m - 1]
            if s_all[i] == 0:
                norm = sqrt(M_PI * r * r / en  *
                             (1 - n * n / (up * up)) * jnup * jnup)
                As[i] = 1.0 / norm
                Bs[i] = 0.0
                Ys[i] = hs[i] / w
            else:
                norm = sqrt(M_PI * r * r / en * jnpup * jnpup)
                As[i] = 0.0
                Bs[i] = 1.0 / norm
                Ys[i] = e1 * w / hs[i]
            us[i] = up
            vs[i] = (1 - 1j) * csqrt(0.5j * (- e2 * w * w + hs[i] * hs[i])) * r
        return us_array, vs_array, As_array, Bs_array, Ys_array
    coefs_C(
        &hs[0], w, &s_all[0], &n_all[0], &m_all[0], num_n_all,
        r, e1, e2, &As[0], &Bs[0])
    for i in range(num_n_all):
        s = s_all[i]
        n = n_all[i]
        en = 1 if n == 0 else 2
        h = hs[i]
        u = (1 + 1j) * csqrt(-0.5j * (e1 * w * w - h * h)) * r
        jv_jvp(n, u, vals)
        jnu = vals[0]
        jnpu = vals[1]
        v = (1 - 1j) * csqrt(0.5j * (- e2 * w * w + h * h)) * r
        kv_kvp(n, v, vals)
        knv = vals[0]
        knpv = vals[1]
        val_u = 2 * M_PI * r * r / en
        val_v = val_u * ((u * jnu) / (v * knv)) ** 2
        ud = upart_diag(n, u, jnu, jnpu, u, jnu, jnpu)
        vd = vpart_diag(n, v, knv, knpv, v, knv, knpv)
        uod = upart_off(n, u, jnu, u, jnu)
        vod = vpart_off(n, v, knv, v, knv)
        a = As[i]
        b = Bs[i]
        Ys[i] = (val_u * (h / w * a * (a * ud + b * uod) +
                          e1 * w / h * b * (b * ud + a * uod)) -
                 val_v * (h / w * a * (a * vd + b * vod) +
                          e2 * w / h * b * (b * vd + a * vod)))
        us[i] = u
        vs[i] = v
    return us_array, vs_array, As_array, Bs_array, Ys_array


@cython.cdivision(True)
cdef cdouble u_func(cdouble h2, cdouble w, cdouble e1, double r):
    return (1 + 1j) * csqrt(-0.5j * (e1 * w ** 2 - h2)) * r


@cython.cdivision(True)
cdef cdouble v_func(cdouble h2, cdouble w, cdouble e2, double r):
    return (1 - 1j) * csqrt(0.5j * (-e2 * w ** 2 + h2)) * r


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def eig_eq_cython(cdouble h2, cdouble w, str pol, int n, cdouble e1, cdouble e2, double r):
    """Return the left value of the characteristic equation"""

    cdef:
        cdouble u = u_func(h2, w, e1, r)
        cdouble v = v_func(h2, w, e2, r)
        cdouble jus, jpus
        cdouble kvs, kpvs
        cdouble te, tm, val
        cdouble vals[2]

    jv_jvp(n, u, vals)
    jus = vals[0]
    jpus = vals[1]
    kv_kvp(n, v, vals)
    kvs = vals[0]
    kpvs = vals[1]
    te = jpus / u + kpvs * jus / (v * kvs)
    tm = e1 * jpus / u + e2 * kpvs * jus / (v * kvs)
    if n == 0:
        if pol == "M":
            val = tm
        else:
            val = te
    else:
        val = (
            tm * te - h2 * (n / w) ** 2 * ((1 / u ** 2 + 1 / v ** 2) * jus) ** 2
        )
    return val


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def jac_cython(double[::1] h2vec, cdouble w, str pol, int n, cdouble e1, cdouble e2, double r):
    """Return Jacobian of the characteristic equation"""

    cdef:
        cdouble h2 = h2vec[0] + 1j * h2vec[1]
        cdouble u = u_func(h2, w, e1, r)
        cdouble v = v_func(h2, w, e2, r)
        cdouble jus, jpus
        cdouble kvs, kpvs, k_k
        cdouble te, tm, val
        cdouble du_dh2, dv_dh2, dte_du, dte_dv, dtm_du, dtm_dv, dre_dh2
        cdouble vals[2]

    jv_jvp(n, u, vals)
    jus = vals[0]
    jpus = vals[1]
    kv_kvp(n, v, vals)
    kvs = vals[0]
    kpvs = vals[1]
    k_k = kpvs / kvs
    te = jpus / u + k_k * jus / v
    tm = e1 * jpus / u + e2 * k_k * jus / v

    du_dh2 = -r / (2 * u)
    dv_dh2 = r / (2 * v)

    dte_du = -(1 - n ** 2 / u ** 2) * jus / u - 2 * jpus / u ** 2 + k_k * jpus / v

    dte_dv = jus * (n ** 2 / v + v * (1 - k_k ** 2) - 2 * k_k) / v ** 2

    dtm_du = e1 * dte_du
    dtm_dv = e2 * dte_dv


    if n == 0:
        if pol == "M":
            val = dtm_du * du_dh2 + dtm_dv * dv_dh2
        else:
            val = dte_du * du_dh2 + dte_dv * dv_dh2
    else:
        dre_dh2 = (
            -((n / w) ** 2)
            * jus
            * (
                jus
                * (
                    (1 / u ** 2 + 1 / v ** 2) ** 2
                    - r * h2 * (1 / u ** 4 - 1 / v ** 4)
                )
                + jpus * 2 * h2 * (1 / u ** 2 + 1 / v ** 2) ** 2
            )
        )
        val = (
            (dte_du * du_dh2 + dte_dv * dv_dh2) * tm
            + (dtm_du * du_dh2 + dtm_dv * dv_dh2) * te
            + dre_dh2
        )
    return np.array([[val.real, -val.imag], [val.imag, val.real]])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def func_cython(
    double[::1] h2vec, cdouble w, str pol, int n, cdouble e1, cdouble e2, double r, cdouble[::1] roots
):
    """Return the value of the characteristic equation

    Args:
        h2: The square of the phase constant.
        w: The angular frequency
        pol: The polarization
        n: The order of the modes
        e1: The permittivity of the core
        e2: The permittivity of the clad.
        r: The radius of hole
        roots: Already obtained roots
    Returns:
        val: A complex indicating the left-hand value of the characteristic
            equation.
    """
    cdef:
        int i
        cdouble h2 = h2vec[0] + 1j * h2vec[1]
        cdouble u = u_func(h2, w, e1, r)
        cdouble v = v_func(h2, w, e2, r)
        cdouble jus, jpus
        cdouble kvs, kpvs
        cdouble te, tm
        cdouble vals[2]
        cdouble f
        cdouble denom = 1.0 + 0.0j
        double norm
        int num = len(roots)

    jv_jvp(n, u, vals)
    jus = vals[0]
    jpus = vals[1]
    kv_kvp(n, v, vals)
    kvs = vals[0]
    kpvs = vals[1]
    te = jpus / u + kpvs * jus / (v * kvs)
    tm = e1 * jpus / u + e2 * kpvs * jus / (v * kvs)
    if n == 0:
        if pol == "M":
            f = tm
        else:
            f = te
    else:
        f = (
            tm * te - h2 * (n / w) ** 2 * ((1 / u ** 2 + 1 / v ** 2) * jus) ** 2
        )
    for i in range(num):
        denom *= h2 - roots[i]
    norm = cabs(denom)
    if norm < 1e-14:
        denom /= norm * 1e14
    f /= denom
    return f.real, f.imag



# -*- coding: utf-8 -*-
# cython: profile=False
from scipy.special.cython_special cimport jv, kv, yv

from . cimport cdouble


cdef double jn(int n, double z) nogil
cdef double jnp(int n, double z) nogil
cdef void jn_jnp(int n, double z, double vals[2]) nogil
cdef void jn_jnp_jnpp(int n, double z, double vals[3]) nogil

cdef cdouble jvp(int n, cdouble z) nogil
cdef void jv_jvp(int n, cdouble z, cdouble vals[2]) nogil
cdef void jv_jvp_jvpp(int n, cdouble z, cdouble vals[3]) nogil

cdef double yn(int n, double z) nogil
cdef double ynp(int n, double z) nogil
cdef void yn_ynp(int n, double z, double vals[2]) nogil
cdef void yn_ynp_ynpp(int n, double z, double vals[3]) nogil

cdef cdouble yvp(int n, cdouble z) nogil
cdef void yv_yvp(int n, cdouble z, cdouble vals[2]) nogil
cdef void yv_yvp_yvpp(int n, cdouble z, cdouble vals[3]) nogil

cdef double knp(int n, double z) nogil
cdef void kn_knp(int n, double z, double vals[2]) nogil

cdef cdouble kvp(int n, cdouble z) nogil
cdef void kv_kvp(int n, cdouble z, cdouble vals[2]) nogil
